/*
 * Copyright (C) 2005 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "para.h"
#include <ortp/ortp.h>
#include <vorbis/vorbisfile.h>
#include "ovrecv.cmdline.h"

#define CHUNK_SIZE 30000

extern gint msg_to_buf(mblk_t *, char *, gint);


static RtpSession *session;
static struct gengetopt_args_info conf;
static ogg_sync_state *sync_out;
static FILE *outfile;
static char *prebuffer;
static int prebuflen;

enum {END_OF_FILE, BEGIN_OF_FILE, OV_HEADER};

static void clog(int ll, char* fmt,...)
{
	va_list argp;

	/* ignore log message if loglevel is not high enough */
	if (ll < conf.loglevel_arg)
		return;
	va_start(argp, fmt);
	vfprintf(stderr, fmt, argp);
	va_end(argp);
}

static void sync_init(void)
{
	if (!sync_out) {
		sync_out = para_malloc(sizeof(ogg_sync_state));
		ogg_sync_init(sync_out);
	}
}

static FILE *open_out_file(char *arg)
{
	FILE *f;

	if (!strcmp(arg, "-")) {
		clog(INFO, "using stdout\n");
		f = stdout;
		goto out;
	}
	clog(INFO, "opening out file\n");
	if (*arg == '|')
		f = popen(arg + 1, "w");
	else
		f = fopen(arg, "w");
	if (!f) {
		clog(EMERG, "cannot open file for writing\n");
		exit(EXIT_FAILURE);
	}
out:
	setvbuf(f, NULL, _IONBF, 0);
	sync_init();
	return f;
}

static void close_out_file(FILE *f, char *arg)
{
	if (!strcmp(arg, "-")) {
		clog(INFO, "NOT closing stdout\n");
		return;
	}
	clog(INFO, "closing out file\n");
	if (*arg == '|') {
		pclose(f);
		return;
	}
	fclose(f);
	if (sync_out) {
		ogg_sync_destroy(sync_out);
		sync_out = NULL;	
	}
}

static int buffer_page(const ogg_page *p)
{
	if (prebuflen + p->header_len + p->body_len >= conf.prebuffer_arg)
		return prebuflen;
	if (!prebuffer)
		prebuffer = para_malloc(conf.prebuffer_arg * sizeof(char));
	memcpy(prebuffer + prebuflen, p->header, p->header_len);
	prebuflen += p->header_len;
	memcpy(prebuffer + prebuflen, p->body, p->body_len);
	prebuflen += p->body_len;
	clog(INFO, "prebuffer: %d\n", prebuflen);
	return -1;

}

static int write_buffer(mblk_t *mp, FILE *outfile)
{	
	int ret, sent_something = 0;
	char *buf;
	ogg_page page;
	static int prebuffering = 1;

	sync_init();
	buf = ogg_sync_buffer(sync_out, CHUNK_SIZE);
	ret = msg_to_buf(mp, buf, CHUNK_SIZE);

	if (ret <= 0)
		goto out;
	ogg_sync_wrote(sync_out, ret);
again:
	ret = -1;
	while (ret < 0) {
		ret = ogg_sync_pageout(sync_out, &page);
		if (ret < 0)
			clog(WARNING, "pageout failed\n");
	}
	if (!ret) {
		clog(DEBUG, "no more pages left\n");
		ret = sent_something;
		goto out;
	}
	if (prebuffering) {
		if (buffer_page(&page) < 0)
			goto again; /* page did fit into prebuffer */
		if (!outfile) {
			prebuflen = 0;
			return 0;
		}
		ret = write(fileno(outfile), prebuffer, prebuflen);
		clog(INFO, "prebuffer write: %d\n", ret);
		prebuffering = 0;
		free(prebuffer);
	}
	clog(DEBUG, "sending page\n");
	sent_something = 1;
	ret = write(fileno(outfile), page.header, page.header_len);
	if (ret < 0) {
		clog(ERROR, "header write: %d\n", ret);
		goto out;
	}
	if (ret < page.header_len) {
		clog(WARNING, "short header write\n");
		ret = -1;
		goto out;
	}
	ret = write(fileno(outfile), page.body, page.body_len);
	if (ret < 0) {
		clog(WARNING, "body write: %d\n", ret);
		goto out;
	}
	if (ret < page.body_len) {
		ret = -1;
		clog(WARNING, "short body write\n");
		goto out;
	}
//	fflush(outfile);
	goto again;
out:	
	return ret;

}

static int handle_other_package(mblk_t *mp, unsigned *time, int have_header)
{
	char buf[CHUNK_SIZE];
	int ret = -1, code, len = msg_to_buf(mp, buf, CHUNK_SIZE);

	if (len <= 0) {
		clog(ERROR, "msg_to_buf failed: %i\n", len);
		return -1;
	}
	code = buf[0];
	clog(INFO, "rcvd ctrl pkg #%i (%d bytes)\n", code, len);
	switch (code) {
	case END_OF_FILE:
		ret = -1;
		if (outfile == stdout)
			goto out;
		if (outfile) {
			close_out_file(outfile, conf.file_arg);
			outfile = NULL;
		}
		*time = 100;
		break;
	case BEGIN_OF_FILE:
		if (outfile == stdout)
			goto out;
		if (outfile)
			close_out_file(outfile, conf.file_arg);
		outfile = open_out_file(conf.file_arg);
		*time = 500;
		break;
	case OV_HEADER:
		if (outfile)
			break;
		*time = 0;
		outfile = open_out_file(conf.file_arg);
		if (have_header)
			goto out;
		ret = write(fileno(outfile), buf + 1, len - 1);
		clog(DEBUG, "wrote header: %d/%d\n", ret, len - 1);
		if (ret < len - 1)
			goto out;
		break;
	default:
		clog(WARNING, "rcvd unknown header pkg\n");
		goto out;
	}
	ret = 1;
out:
	freemsg(mp);
	return ret;
}

int main(int argc, char *argv[])
{
	int ret;
	int have_header = 0;
	unsigned time = 10;
	guint32 ts = 0;
	mblk_t *mp;

	cmdline_parser(argc, argv, &conf);
	ortp_init();
	ortp_scheduler_init();
	ortp_set_debug_file("oRTP", NULL);
	session = rtp_session_new(RTP_SESSION_RECVONLY);
	rtp_session_set_scheduling_mode(session, 1);
	rtp_session_set_blocking_mode(session, 1);
	rtp_session_set_local_addr(session, conf.hostname_arg, conf.port_arg);
	rtp_session_set_payload_type(session, PAYLOAD_AUDIO_CONTINUOUS);
recv:
	mp = rtp_session_recvm_with_ts(session, ts);
	if (!mp) {
		clog(DEBUG, "did not receive buffer, time = %i\n", time);
		if (time < 1000)
			time += 100;
		ts += time;
		goto recv;
	}
	if (rtp_session_get_payload_type(session) == PAYLOAD_OTHER) {
		ret = handle_other_package(mp, &time, have_header);
		if (ret < 0)
			goto out;
		ts += 100;
		goto recv;
	}
	if (!outfile && !have_header) {
		clog(NOTICE, "have data, but no header, prebuffering...\n");
		write_buffer(mp, NULL);
		goto recv;
	}
	clog(DEBUG, "have data pkg\n");
	if (!outfile)
		outfile = open_out_file(conf.file_arg);
	ret = write_buffer(mp, outfile);
	freemsg(mp);
	if (ret < 0)
		goto out;
	time = 100;
	ts += time;
	goto recv;
out:
	clog(NOTICE, "song finished, exit code: %d\n", ret);
	if (outfile) {
		close_out_file(outfile, conf.file_arg);
		outfile = NULL;
	}
	rtp_session_destroy(session);
	ortp_exit();
	ortp_global_stats_display();
	return ret;
}
